"""21cmEMU: An emulator of 21cmFAST summaries."""

__version__ = "1.0.8"
from .emulator import Emulator
from .get_emulator import get_emu_data
from .inputs import EmulatorInput
from .outputs import EmulatorOutput
from .outputs import RawEmulatorOutput
from .properties import COSMO_PARAMS
from .properties import FLAG_OPTIONS
from .properties import USER_PARAMS
from .properties import EmulatorProperties



"""Module that interacts with the Emulator Tensorflow model."""
from __future__ import annotations

import logging
from typing import Any

import numpy as np
import tensorflow as tf

from .config import CONFIG
from .get_emulator import get_emu_data
from .inputs import EmulatorInput
from .inputs import ParamVecType
from .outputs import EmulatorOutput
from .outputs import RawEmulatorOutput
from .properties import emulator_properties


log = logging.getLogger(__name__)


class Emulator:
    r"""A class that loads an emulator and uses it to obtain 21cmFAST summaries.

    Parameters
    ----------
    version : str, optional
        Emulator version to use/download, default is 'latest'.
    """

    def __init__(self, version: str = "latest"):
        get_emu_data(version=version)

        emu = tf.keras.models.load_model(CONFIG.emu_path, compile=False)

        self.model = emu
        self.inputs = EmulatorInput()
        self.properties = emulator_properties

    def __getattr__(self, name: str) -> Any:
        """Allow access to emulator properties directly from the emulator object."""
        return getattr(self.properties, name)

    def predict(
        self, astro_params: ParamVecType, verbose: bool = False
    ) -> tuple[np.ndarray, EmulatorOutput, dict[str, np.ndarray]]:
        r"""Call the emulator, evaluate it at the given parameters, restore dimensions.

        Parameters
        ----------
        astro_params : np.ndarray or dict
            An array with the nine astro_params input all $\in [0,1]$ OR in the
            21cmFAST AstroParams input units. Dicts (e.g. p21.AstroParams.defining_dict)
            are also accepted formats. Arrays of only dicts are accepted as well
            (for batch evaluation).
        verbose : bool, optional
            If True, prints the emulator prediction.

        Returns
        -------
        theta : np.ndarray
            The normalized parameters used to evaluate the emulator.
        emu : EmulatorOutput
            The emulator output, with dimensions restored.
        errors : dict
            The mean error on the test set (i.e. independent of theta).
        """
        theta = self.inputs.make_param_array(astro_params, normed=True)
        emu = RawEmulatorOutput(self.model.predict(theta, verbose=verbose))
        emu = emu.get_renormalized()

        errors = self.get_errors(emu, theta)

        return theta, emu, errors

    def get_errors(
        self, emu: EmulatorOutput, theta: np.ndarray | None = None
    ) -> dict[str, np.ndarray]:
        """Calculate the emulator error on its outputs.

        Parameters
        ----------
        emu : dict
            Dict containing the emulator predictions, defined in Emulator.predict
        theta : dict
            Dict containing the normalized parameters, also defined in Emulator.predict

        Returns
        -------
        The mean error on the test set (i.e. independent of theta) with all units
        restored and logs removed.
        """
        # For now, we return the mean emulator error (obtained from the test set) for
        # each summary. All errors are the median absolute difference between test set
        # and prediction AFTER units have been restored AND log has been removed.
        return {
            "PS_err": self.PS_err,
            "Tb_err": self.Tb_err,
            "xHI_err": self.xHI_err,
            "Ts_err": self.Ts_err,
            "UVLFs_err": self.UVLFs_err,
            "UVLFs_logerr": self.UVLFs_logerr,
            "tau_err": self.tau_err,
        }



"""Module containing functionality for handling emulator inputs."""
from __future__ import annotations

from typing import Dict
from typing import Sequence
from typing import Union

import numpy as np

from .properties import emulator_properties as properties


SingleParamVecType = Union[Dict[str, float], np.ndarray, Sequence[float]]
ParamVecType = Union[Sequence[SingleParamVecType], SingleParamVecType]


class EmulatorInput:
    """Class for handling emulator inputs."""

    astro_param_keys = (
        "F_STAR10",
        "ALPHA_STAR",
        "F_ESC10",
        "ALPHA_ESC",
        "M_TURN",
        "t_STAR",
        "L_X",
        "NU_X_THRESH",
        "X_RAY_SPEC_INDEX",
    )

    def _format_single_theta_vector(self, theta: SingleParamVecType) -> np.ndarray:
        if len(theta) != len(self.astro_param_keys):
            raise ValueError(
                "One of the parameter vectors given is not the correct length. Got "
                f"{len(theta)} but require {len(self.astro_param_keys)}."
            )

        if isinstance(theta, dict):
            return np.array([float(theta[key]) for key in self.astro_param_keys])
        elif isinstance(theta, np.ndarray):
            return theta.astype(float)
        elif isinstance(theta, list):
            return np.array(theta, dtype=float)
        else:
            raise TypeError(
                "astro_params is in the wrong format. Should be a dict of astro params,"
                " list of astro params (in same order as astro_param_keys), or an array"
                " of astro params (in same order as astro_param_keys), OR a sequence of"
                " such."
            )

    def make_param_array(
        self,
        astro_params: ParamVecType,
        normed: bool = True,
    ) -> np.ndarray:
        """Format the astro_params input to be a numpy array.

        Parameters
        ----------
        astro_params : dict or np.ndarray
            Input parameters to the emulator, to be reformatted as a numpy array
            compatible with the emulator. Can be a dict of astro params or a list/array
            of floats, in the same order as astro_param_keys. It could also be a list
            or array of such objects, for batch evaluation (the type can be mixed).
        normed : bool, optional
            Whether to return the parameters normalized or not (i.e. between 0 and 1).
        """
        if not hasattr(astro_params, "__len__"):
            raise TypeError(
                "astro_params is in the wrong format. Should be a dict of astro params,"
                " list of astro params (in same order as astro_param_keys), or an array"
                " of astro params (in same order as astro_param_keys), OR a sequence of"
                " such."
            )
        if isinstance(astro_params, dict):
            astro_params = [astro_params]
        else:
            if not hasattr(astro_params[0], "__len__"):
                astro_params = [astro_params]

        theta = np.array(
            [self._format_single_theta_vector(theta) for theta in astro_params]
        )

        params_normed = theta.min() >= 0 and theta.max() <= 1
        if (params_normed and normed) or (not params_normed and not normed):
            return theta
        elif params_normed:
            return self.undo_normalization(theta)
        else:
            return self.normalize(theta)

    def make_list_of_dicts(
        self, theta: ParamVecType, normed: bool = True
    ) -> list[dict[str, float]]:
        """Make a list of dicts from a theta array.

        Parameters
        ----------
        theta : np.ndarray
            Input parameters, in any format accepted by :func:`~make_param_array`.
        normed : bool, optional
            Whether to return the parameters normalized or not (i.e. between 0 and 1).

        Returns
        -------
        list of dicts
            List of dicts of astro params, one for each parameter set.
        """
        theta = self.make_param_array(theta, normed=normed)
        return [dict(zip(self.astro_param_keys, theta[i])) for i in range(len(theta))]

    def normalize(self, theta: np.ndarray) -> np.ndarray:
        """Normalize the parameters.

        Parameters
        ----------
        theta : np.ndarray
            Input parameters, strictly in 2D array format, with shape
            (n_batch, n_params).

        Returns
        -------
        np.ndarray
            Normalized parameters, with shape (n_batch, n_params).
        """
        theta_woutdims = theta.copy()
        theta_woutdims[:, 7] /= 1000
        theta_woutdims -= properties.limits[:, 0]
        theta_woutdims /= properties.limits[:, 1] - properties.limits[:, 0]
        return theta_woutdims

    def undo_normalization(self, theta: np.ndarray) -> np.ndarray:
        """Undo the normalization of the parameters.

        Parameters
        ----------
        theta : np.ndarray
            Input parameters, strictly in 2D array format, with shape
            (n_batch, n_params).

        Returns
        -------
        np.ndarray
            Un-normalized parameters, with shape (n_batch, n_params).
        """
        theta_wdims = theta.copy()
        theta_wdims *= properties.limits[:, 1] - properties.limits[:, 0]
        theta_wdims += properties.limits[:, 0]
        theta_wdims[:, 7] *= 1000
        return theta_wdims



"""Module whose functionality is to organise the emulator output."""
from __future__ import annotations

import dataclasses as dc
from dataclasses import dataclass
from pathlib import Path
from typing import Generator

import numpy as np

from .properties import emulator_properties


@dataclass(frozen=True)
class EmulatorOutput:
    """A simple class that makes it easier to access the corrected emulator output."""

    Tb: np.ndarray
    xHI: np.ndarray
    Ts: np.ndarray
    PS: np.ndarray
    tau: np.ndarray
    UVLFs: np.ndarray

    properties = emulator_properties

    def keys(self) -> Generator[str, None, None]:
        """Yield the keys of the main data products."""
        for k in dc.fields(self):
            yield k.name

    def items(self) -> Generator[tuple[str, np.ndarray], None, None]:
        """Yield the keys and values of the main data products, like a dict."""
        for k in self.keys():
            yield k, getattr(self, k)

    def __getitem__(self, key: str) -> np.ndarray:
        """Allow access to attributes as items."""
        return getattr(self, key)

    @property
    def k(self) -> np.ndarray:
        """The k-values of the power spectra."""
        return self.properties.ks_cut

    @property
    def Muv(self) -> np.ndarray:
        """The Muv-values of the UVLFs."""
        # Crop the M_UV to -20 to -10 range
        m = np.logical_and(
            self.properties.UVLFs_MUVs <= -10, self.properties.UVLFs_MUVs >= -20
        )
        return self.properties.UVLFs_MUVs[m]

    @property
    def UVLF_redshifts(self) -> np.ndarray:
        """The redshifts of the UVLFs."""
        return self.properties.uv_lf_zs

    @property
    def PS_redshifts(self) -> np.ndarray:
        """The redshifts of the power spectra."""
        return self.properties.zs_cut

    @property
    def redshifts(self) -> np.ndarray:
        """The redshifts of all quantities except the PS."""
        return self.properties.zs

    def squeeze(self):
        """Return a new EmulatorOutput with all dimensions of length 1 removed."""
        return EmulatorOutput(**{k: np.squeeze(v) for k, v in self.items()})

    def write(
        self,
        fname: str | Path,
        theta: np.ndarray | dict | None = None,
        store: list[str] | None = None,
        clobber: bool = False,
    ):
        """Write this instance's data to a file.

        This saves the output as a numpy .npz file. The output is saved as a dictionary
        with the keys being the names of the attributes of this class and the values
        being the corresponding values of those attributes. If theta is not None, then
        the inputs are also saved under the key "inputs".

        Parameters
        ----------
        fname : str or Path
            The filename to write to.
        theta : np.ndarray or dict or None, optional
            The input parameters associated with this output data to write to the file.
            If None, the inputs are not written.
        store : list of str or None, optional
            The names of the attributes to write to the file. If None, all attributes
            are written.
        clobber : bool, optional
            Whether to overwrite the file if it already exists.
        """
        if store is None:
            store = list(self.__dict__.keys())

        pth = Path(fname)
        if pth.exists() and not clobber:
            raise ValueError(f"File {pth} exists and clobber=False.")

        out = {k: getattr(self, k) for k in store}
        if theta is not None:
            out["inputs"] = theta

        np.savez(fname, out)


@dataclass(frozen=True)
class RawEmulatorOutput:
    """A simple data-class that makes it easier to access the raw emulator output.

    Parameters
    ----------
    output : np.ndarray
        The raw output array from the emulator.
    """

    output: np.ndarray

    properties = emulator_properties

    @property
    def nz(self) -> int:
        """Number of redshifts in the output."""
        return np.array(self.properties.zs).shape[0]

    @property
    def nparams(self) -> int:
        """Number of sets of parameters in the output."""
        return self.output.shape[0]

    @property
    def Tb(self) -> np.ndarray:
        """Mean 21cm brightness temperature in mK as a function of redshift."""
        return self.output[:, : self.nz]

    @property
    def xHI(self) -> np.ndarray:
        """Neutral fraction as a function of redshift."""
        return self.output[:, self.nz : 2 * self.nz]

    @property
    def Ts(self) -> np.ndarray:
        """Mean spin temperature in K as a function of redshift."""
        return self.output[:, 2 * self.nz : 3 * self.nz]

    @property
    def reshift_where_Ts_becomes_defined(self) -> np.ndarray:
        """The redshift at which the spin temperature becomes well-defined."""
        return self.output[:, 3 * self.nz]

    @property
    def PS(self) -> np.ndarray:
        """The power spectrum in mK^2 as a function of redshift and k."""
        return self.output[:, 3 * self.nz + 1 : 3 * self.nz + 1 + 60 * 12].reshape(
            (-1, 60, 12)
        )

    @property
    def tau(self) -> np.ndarray:
        """The optical depth of reionization."""
        return self.output[:, self.nz * 3 + 1 + 60 * 12]

    @property
    def UVLFs(self) -> np.ndarray:
        """The UV luminosity functions as a function of z and Muv."""
        full_UVLFs = self.output[:, self.nz * 3 + 1 + 60 * 12 + 1 :].reshape(
            (-1, len(self.properties.uv_lf_zs), len(self.properties.UVLFs_MUVs))
        )
        # Crop M_UVs to allowed range of [-20,-10]
        m = np.logical_and(
            self.properties.UVLFs_MUVs <= -10, self.properties.UVLFs_MUVs >= -20
        )
        return full_UVLFs[..., m]

    def renormalize(self, name: str):
        """Renormalize a normalized quantity.

        This ajudsts the quantity (as it exists in this class) back to its native
        range by adding the emulator data mean and multiplying by the emulator data
        standard deviation.
        """
        if name not in self.properties.normalized_quantities:
            raise ValueError(
                f"Cannot renormalize {name}. It is not a normalized quantity."
            )
        return getattr(self.properties, f"{name}_mean") + getattr(
            self.properties, f"{name}_std"
        ) * getattr(self, name)

    def get_renormalized(self) -> EmulatorOutput:
        """Get the output with normalized quantities re-normalized.

        Returns
        -------
        EmulatorOutput
            The emulator output with normalized quantities re-normalized back to
            physical units. Nothing is in log except UV LFs.
        """
        # Restore dimensions
        # Renormalize stuff that needs renormalization
        renorm = {k: self.renormalize(k) for k in self.properties.normalized_quantities}

        other = {
            k.name: getattr(self, k.name)
            for k in dc.fields(EmulatorOutput)
            if k.name not in renorm
        }

        out = {**renorm, **other}

        # Set the xHI < z(Ts undefined) to 0
        # For Ts, set it to NaN
        for i in range(self.nparams):
            zbin = np.argmin(
                abs(self.properties.zs - self.reshift_where_Ts_becomes_defined[i])
            )
            if out["xHI"][i, zbin] < 1e-1:
                out["xHI"][i, :zbin] = 0.0
            out["Ts"][i, :zbin] = np.nan

        # Undo log10 on some quantities
        out["PS"] = 10 ** out["PS"]
        out["Ts"] = 10 ** out["Ts"]
        out["tau"] = 10 ** out["tau"]

        return EmulatorOutput(**out).squeeze()



"""Download and install the emulator data."""
from __future__ import annotations

import logging
from warnings import warn

import git

from .config import CONFIG


log = logging.getLogger(__name__)


def get_emu_data(version: str = "latest"):
    """Download a version of the 21cmEMU emulator.

    Parameters
    ----------
    version : str, optional
        When multiple versions will be available, one will be able to specify
        the version number instead of the link. Default is 'latest'.
    """
    if (CONFIG.data_path / "21cmEMU").exists():
        repo = git.Repo(CONFIG.data_path / "21cmEMU")
    elif not CONFIG["disable-network"]:
        URL = "https://huggingface.co/DanielaBreitman/21cmEMU"
        repo = git.Repo.clone_from(URL, CONFIG.data_path / "21cmEMU")

    # Check download
    p = CONFIG.data_path / "21cmEMU" / "21cmEMU" / "saved_model.pb"
    if not p.exists() or p.stat().st_size < 1e6:
        raise RuntimeError(
            "The emulator huggingface repo was not cloned properly.\n"
            "Check that git-lfs is installed properly on your system.\n"
            "If git-lfs cannot be installed or internet "
            "connection is not available, "
            "manually clone the repo on another machine with git-lfs"
            "and internet using\n"
            "git clone -v -- https://huggingface.co/DanielaBreitman/21cmEMU\n"
            "Then, ensure that it downloaded fully by running: "
            "du -sh 21cmEMU \n"
            "The folder should be about 500M. "
            "Now copy this folder and its contents "
            "over to your other machine and put it in "
            " ~/.local/share/py21cmEMU/21cmEMU "
        )

    # Pull latest changes
    try:
        if CONFIG["disable-network"]:
            raise RuntimeError("Network is disabled via config")

        repo.remotes.origin.pull()
    except Exception as e:
        warn(f"Skipping the pulling step. Error received: {e}", stacklevel=2)

    versions = sorted(
        [tag.name.lower() for tag in repo.tags if tag.name.lower().startswith("v")]
    )

    if version == "latest":
        version = repo.git.checkout("main")
    elif version.lower() in versions:
        repo.git.checkout(version)
    else:
        raise ValueError(
            f"Version {version} not available. Must be one of {versions}. "
        )



"""A module definining the static properties of the Emulator."""
from __future__ import annotations

from pathlib import Path

import numpy as np


USER_PARAMS = {
    "BOX_LEN": 250,
    "DIM": 512,
    "HII_DIM": 128,
    "USE_FFTW_WISDOM": True,
    "HMF": 1,
    "USE_RELATIVE_VELOCITIES": False,
    "POWER_SPECTRUM": 0,
    "N_THREADS": 1,
    "PERTURB_ON_HIGH_RES": False,
    "NO_RNG": False,
    "USE_INTERPOLATION_TABLES": True,
    "FAST_FCOLL_TABLES": False,
    "USE_2LPT": True,
    "MINIMIZE_MEMORY": False,
}

COSMO_PARAMS = {
    "SIGMA_8": 0.82,
    "hlittle": 0.6774,
    "OMm": 0.3075,
    "OMb": 0.0486,
    "POWER_INDEX": 0.97,
}

FLAG_OPTIONS = {
    "USE_HALO_FIELD": False,
    "USE_MINI_HALOS": False,
    "USE_MASS_DEPENDENT_ZETA": True,
    "SUBCELL_RSD": True,
    "INHOMO_RECO": True,
    "USE_TS_FLUCT": True,
    "M_MIN_in_Mass": False,
    "PHOTON_CONS": True,
    "FIX_VCB_AVG": False,
}


class EmulatorProperties:
    """A class that contains the properties of the emulator."""

    def __init__(self):
        here = Path(__file__).parent
        all_emulator_numbers = np.load(here / "emulator_constants.npz")
        self._data = all_emulator_numbers

        self.zs = all_emulator_numbers["zs"]
        self.limits = all_emulator_numbers["limits"]
        self.zs_cut = self.zs[:60]
        self.ks_cut = all_emulator_numbers["ks"][1:-3]
        self.PS_mean = all_emulator_numbers["PS_mean"]
        self.PS_std = all_emulator_numbers["PS_std"]
        self.Tb_mean = all_emulator_numbers["Tb_mean"]
        self.Tb_std = all_emulator_numbers["Tb_std"]
        self.Ts_mean = all_emulator_numbers["Ts_mean"]
        self.Ts_std = all_emulator_numbers["Ts_std"]

        self.tau_mean = all_emulator_numbers["tau_mean"]
        self.tau_std = all_emulator_numbers["tau_std"]

        self.UVLFs_mean = all_emulator_numbers["UVLFs_mean"]
        self.UVLFs_std = all_emulator_numbers["UVLFs_std"]
        self.UVLFs_MUVs = np.append(np.arange(-25, -15, 1.0), np.arange(-15, -4.5, 0.5))
        self.uv_lf_zs = np.array([6, 7, 8, 10])

        self.PS_err = all_emulator_numbers["PS_err"]
        self.Tb_err = all_emulator_numbers["Tb_err"]
        self.Ts_err = all_emulator_numbers["Ts_err"]
        self.xHI_err = all_emulator_numbers["xHI_err"]
        self.tau_err = all_emulator_numbers["tau_err"]
        self.UVLFs_err = all_emulator_numbers["UVLFs_err"]
        self.UVLFs_logerr = all_emulator_numbers["UVLFs_logerr"]

        self.flag_options = FLAG_OPTIONS
        self.user_params = USER_PARAMS
        self.cosmo_params = COSMO_PARAMS

    @property
    def normalized_quantities(self) -> list[str]:
        """Return a list of the normalized quantities predicted by the emulator."""
        return [k.split("_")[0] for k in self._data if k.endswith("_mean")]


emulator_properties = EmulatorProperties()



"""User-facing configuration for py21cmEMU."""
from __future__ import annotations

import logging
from contextlib import contextmanager
from pathlib import Path
from typing import Any
from typing import Generator

import toml
from appdirs import AppDirs


log = logging.getLogger(__name__)

APPDIR = AppDirs("py21cmEMU")


class Config:
    """Class that handles the configuration file."""

    def __init__(self, config_file: str | Path | None = None) -> None:
        if config_file is None:
            config_file = Path(APPDIR.user_config_dir) / "config.toml"
        else:
            config_file = Path(config_file)

        self.config_file = config_file

        if not self.config_file.exists():
            self.config_file.parent.mkdir(parents=True, exist_ok=True)
            self.config_file.touch()
            log.info(f"Created new config file at {self.config_file}")

        self.config = toml.loads(self.config_file.read_text())

        if "data-path" not in self:
            self["data-path"] = APPDIR.user_data_dir

        if not self.data_path.exists():
            self.data_path.mkdir(parents=True, exist_ok=True)

        if "disable-network" not in self:
            self["disable-network"] = False

    @property
    def emu_path(self) -> Path:
        """Get the path to the emulator data."""
        return Path(self["data-path"]) / "21cmEMU" / "21cmEMU"

    @property
    def data_path(self) -> Path:
        """Get the path to the data directory."""
        return Path(self["data-path"])

    def __getitem__(self, key: str) -> Any:
        """Get a value from the config file."""
        return self.config[key]

    def __setitem__(self, key: str, value: Any):
        """Set a value in the config file."""
        self.config[key] = value
        self.config_file.write_text(toml.dumps(self.config))

    def __delitem__(self, key: str) -> None:
        """Delete a value from the config file."""
        del self.config[key]
        self.config_file.write_text(toml.dumps(self.config))

    def __contains__(self, key: str) -> bool:
        """Check if a key is in the config file."""
        return key in self.config

    def __repr__(self) -> str:
        """Get the string representation of the config file."""
        return repr(self.config)

    def __str__(self) -> str:
        """Get the string representation of the config file."""
        return str(self.config)

    def keys(self) -> Generator[str, None, None]:
        """Yield the keys in the config file."""
        yield from self.config.keys()

    def values(self) -> Generator[Any, None, None]:
        """Yield the values in the config file."""
        yield from self.config.values()

    def items(self) -> Generator[tuple[str, Any], None, None]:
        """Yield the keys and values of the main data products, like a dict."""
        yield from self.config.items()

    def update(self, **kw) -> None:
        """Update the config file with new values."""
        self.config.update(**kw)
        self.config_file.write_text(toml.dumps(self.config))

    @contextmanager
    def use(self, **kw) -> Generator[None, None, None]:
        """Use configuration values temporarily."""
        old = {k: self[k] for k in kw}
        self.update(**kw)
        try:
            yield
        finally:
            self.update(**old)


CONFIG = Config()
